////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                            //
//                                       WATTO STUDIOS                                        //
//                             Java Code, Programs, and Software                              //
//                                    http://www.watto.org                                    //
//                                                                                            //
//                           Copyright (C) 2004-2010  WATTO Studios                           //
//                                                                                            //
// This program is free software; you can redistribute it and/or modify it under the terms of //
// the GNU General Public License published by the Free Software Foundation; either version 2 //
// of the License, or (at your option) any later versions. This program is distributed in the //
// hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranties //
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License //
// at http://www.gnu.org for more details. For updates and information about this program, go //
// to the WATTO Studios website at http://www.watto.org or email watto@watto.org . Thanks! :) //
//                                                                                            //
////////////////////////////////////////////////////////////////////////////////////////////////

package org.watto.program.android.sync.steam;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.net.URL;
import java.util.Calendar;
import org.watto.android.Language;
import org.watto.android.R;
import org.watto.program.android.sync.steam.array.ArrayResizer;
import org.watto.program.android.sync.steam.datatype.CustomUrlData;
import org.watto.program.android.sync.steam.datatype.FavoriteGameData;
import org.watto.program.android.sync.steam.datatype.HeadlineData;
import org.watto.program.android.sync.steam.datatype.HoursPlayedData;
import org.watto.program.android.sync.steam.datatype.MemberSinceData;
import org.watto.program.android.sync.steam.datatype.MostPlayedGameData;
import org.watto.program.android.sync.steam.datatype.MostPlayedGameHoursData;
import org.watto.program.android.sync.steam.datatype.OnlineStateData;
import org.watto.program.android.sync.steam.datatype.PrimaryGroupData;
import org.watto.program.android.sync.steam.datatype.StatusData;
import org.watto.program.android.sync.steam.datatype.SteamRatingData;
import org.watto.program.android.sync.steam.datatype.SummaryData;
import org.watto.program.android.sync.steam.datatype.UserIdData;
import org.watto.program.android.sync.steam.datatype.WebsiteData;
import org.watto.program.android.sync.steam.datatype.WebsiteTitleData;
import android.content.ContentProviderOperation;
import android.content.ContentValues;
import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.net.Uri;
import android.provider.ContactsContract;
import android.provider.ContactsContract.CommonDataKinds.GroupMembership;
import android.provider.ContactsContract.CommonDataKinds.Nickname;
import android.provider.ContactsContract.CommonDataKinds.Phone;
import android.provider.ContactsContract.CommonDataKinds.Photo;
import android.provider.ContactsContract.CommonDataKinds.StructuredName;
import android.provider.ContactsContract.CommonDataKinds.StructuredPostal;
import android.provider.ContactsContract.Data;
import android.provider.ContactsContract.Groups;
import android.provider.ContactsContract.RawContacts;


/***********************************************************************************************
Helper class for storing data in the platform content providers
***********************************************************************************************/
public class ContactOperations {

  ContentValues content;
  ContentProviderOperation.Builder operationBuilder;
  BatchOperation batchOperation;
  Context context;
  boolean yield;
  long rawContactId = 0;
  int backReference;
  boolean newContact;


  /***********************************************************************************************
  Constructor
  ***********************************************************************************************/
  public ContactOperations(Context context, BatchOperation batchOperation){
    this.content = new ContentValues();
    this.yield = true;
    this.context = context;
    this.batchOperation = batchOperation;
  }


  /***********************************************************************************************
  Constructor for adding a new User
  ***********************************************************************************************/
  public ContactOperations(Context context, int userId, String accountName, BatchOperation batchOperation){
    this(context,batchOperation);
    this.backReference = batchOperation.size();
    this.newContact = true;

    content.put(RawContacts.SOURCE_ID,userId);
    content.put(RawContacts.ACCOUNT_TYPE,GlobalConstants.ACCOUNT_TYPE);
    content.put(RawContacts.ACCOUNT_NAME,accountName);

    this.operationBuilder = addInsertOperation(RawContacts.CONTENT_URI,true).withValues(content);
    batchOperation.add(operationBuilder.build());
  }


  /***********************************************************************************************
  Constructor for adding a new Group
  ***********************************************************************************************/
  public ContactOperations(Context context, int groupId, String accountName, BatchOperation batchOperation, boolean addingGroup){
    this(context,batchOperation);
    this.backReference = batchOperation.size();
    this.newContact = true;
  }


  /***********************************************************************************************
  Constructor for updating an existing User
  ***********************************************************************************************/
  public ContactOperations(Context context, long rawContactId, BatchOperation batchOperation){
    this(context,batchOperation);
    this.newContact = false;
    this.rawContactId = rawContactId;
  }


  /***********************************************************************************************
  Adds a custom url
  @param customUrl the custom url to add
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addCustomUrl(String customUrl){
    content.clear();
    if (customUrl != null && !customUrl.equals("")) {
      content.put(CustomUrlData.DATA,customUrl);
      content.put(CustomUrlData.LABEL,context.getString(R.string.label_customUrl));
      content.put(Data.MIMETYPE,CustomUrlData.CONTENT_ITEM_TYPE);
      addInsertOperation();
    }
    return this;
  }


  /***********************************************************************************************
  Adds an update operation into the batch
  ***********************************************************************************************/
  public void addDeleteOperation(Uri uri){
    operationBuilder = addDeleteOperation(uri,yield);
    yield = false;
    batchOperation.add(operationBuilder.build());
  }


  /***********************************************************************************************
  Adds the favorite game
  @param favoriteGame the favorite game to add
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addFavoriteGame(String favoriteGame){
    content.clear();
    if (favoriteGame != null && !favoriteGame.equals("")) {
      content.put(FavoriteGameData.DATA,favoriteGame);
      content.put(FavoriteGameData.LABEL,context.getString(R.string.label_favoriteGame));
      content.put(Data.MIMETYPE,FavoriteGameData.CONTENT_ITEM_TYPE);
      addInsertOperation();
    }
    return this;
  }


  /***********************************************************************************************
  Adds a "Steam" group
  @param title the name of the group
  @param notes a description of the group
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addGroup(String title,String notes){

    content.clear();
    if (title != null && !title.equals("") && notes != null && !notes.equals("")) {
      content.put(Groups.ACCOUNT_TYPE,GlobalConstants.ACCOUNT_TYPE);
      content.put(Groups.ACCOUNT_NAME,GlobalConstants.ACCOUNT_NAME);
      content.put(Groups.TITLE,title);
      content.put(Groups.NOTES,notes);
      content.put(Groups.GROUP_VISIBLE,"1");
      content.put(Groups.SHOULD_SYNC,"1");
    }

    if (content.size() > 0) {
      operationBuilder = addInsertOperation(addCallerIsSyncAdapterParameter(Groups.CONTENT_URI),yield);
      operationBuilder.withValues(content);

      yield = false;
      batchOperation.add(operationBuilder.build());
      // MUST execute here, so the group is created and can be referenced when adding users to the group
      batchOperation.execute();

    }

    return this;

  }


  /***********************************************************************************************
  Adds a contact to a group
  @param groupId the ID of the group to add the contact to
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addGroupMembership(long groupId){
    content.clear();
    if (groupId != -1) {
      content.put(Data.MIMETYPE,GroupMembership.CONTENT_ITEM_TYPE);
      content.put(GroupMembership.GROUP_ROW_ID,groupId);
    }

    if (content.size() > 0) {
      addInsertOperation();
    }
    return this;
  }


  /***********************************************************************************************
  Adds the headline
  @param headline the headline to add
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addHeadline(String headline){
    content.clear();
    if (headline != null && !headline.equals("")) {
      content.put(HeadlineData.DATA,headline);
      content.put(HeadlineData.LABEL,context.getString(R.string.label_headline));
      content.put(Data.MIMETYPE,HeadlineData.CONTENT_ITEM_TYPE);
      addInsertOperation();
    }
    return this;
  }


  /***********************************************************************************************
  Adds the hours played
  @param hoursPlayed the hours played to add
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addHoursPlayed(String hoursPlayed){
    content.clear();
    if (hoursPlayed != null && !hoursPlayed.equals("")) {
      content.put(HoursPlayedData.DATA,hoursPlayed);
      content.put(HoursPlayedData.LABEL,context.getString(R.string.label_hoursPlayed));
      content.put(Data.MIMETYPE,HoursPlayedData.CONTENT_ITEM_TYPE);
      addInsertOperation();
    }
    return this;
  }


  /***********************************************************************************************
  Adds an insert operation into the batch
  ***********************************************************************************************/
  public void addInsertOperation(){

    if (!newContact) {
      content.put(Phone.RAW_CONTACT_ID,rawContactId);
    }

    operationBuilder = addInsertOperation(addCallerIsSyncAdapterParameter(Data.CONTENT_URI),yield);
    operationBuilder.withValues(content);

    if (newContact) {
      operationBuilder.withValueBackReference(Data.RAW_CONTACT_ID,backReference);
    }

    yield = false;
    batchOperation.add(operationBuilder.build());
  }


  /***********************************************************************************************
  Adds a location
  @param location the location
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addLocation(String location){
    content.clear();
    if (location != null && !location.equals("")) {
      content.put(StructuredPostal.MIMETYPE,StructuredPostal.CONTENT_ITEM_TYPE);
      content.put(StructuredPostal.FORMATTED_ADDRESS,location);
      content.put(StructuredPostal.TYPE,StructuredPostal.TYPE_CUSTOM);
      content.put(StructuredPostal.LABEL,context.getString(R.string.label_location));

      String[] splitLocation = location.split(", ");
      int splitCount = splitLocation.length;
      if (splitCount == 1) {
        content.put(StructuredPostal.COUNTRY,splitLocation[0]);
      }
      else if (splitCount == 2) {
        content.put(StructuredPostal.REGION,splitLocation[0]);
        content.put(StructuredPostal.COUNTRY,splitLocation[1]);
      }
      else if (splitCount == 3) {
        content.put(StructuredPostal.CITY,splitLocation[0]);
        content.put(StructuredPostal.REGION,splitLocation[1]);
        content.put(StructuredPostal.COUNTRY,splitLocation[2]);
      }
    }

    if (content.size() > 0) {
      addInsertOperation();
    }
    return this;
  }


  /***********************************************************************************************
  Adds a member since data
  @param memberSince the member since data to add
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addMemberSince(String memberSince){
    content.clear();
    if (memberSince != null && !memberSince.equals("")) {
      content.put(MemberSinceData.DATA,memberSince);
      content.put(MemberSinceData.LABEL,context.getString(R.string.label_memberSince));
      content.put(Data.MIMETYPE,MemberSinceData.CONTENT_ITEM_TYPE);
      addInsertOperation();
    }
    return this;
  }


  /***********************************************************************************************
  Adds the most played game
  @param mostPlayedGame the most played game to add
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addMostPlayedGame(String mostPlayedGame){
    content.clear();
    if (mostPlayedGame != null && !mostPlayedGame.equals("")) {
      content.put(MostPlayedGameData.DATA,mostPlayedGame);
      content.put(MostPlayedGameData.LABEL,context.getString(R.string.label_mostPlayedGame));
      content.put(Data.MIMETYPE,MostPlayedGameData.CONTENT_ITEM_TYPE);
      addInsertOperation();
    }
    return this;
  }


  /***********************************************************************************************
  Adds the most played game hours
  @param mostPlayedGameHours the most played game hours to add
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addMostPlayedGameHours(String mostPlayedGameHours){
    content.clear();
    if (mostPlayedGameHours != null && !mostPlayedGameHours.equals("")) {
      content.put(MostPlayedGameHoursData.DATA,mostPlayedGameHours);
      content.put(MostPlayedGameHoursData.LABEL,context.getString(R.string.label_mostPlayedGameHours));
      content.put(Data.MIMETYPE,MostPlayedGameHoursData.CONTENT_ITEM_TYPE);
      addInsertOperation();
    }
    return this;
  }


  /***********************************************************************************************
  Adds the online state
  @param onlineState the online state to add
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addOnlineState(String onlineState){
    content.clear();
    if (onlineState != null && !onlineState.equals("")) {
      content.put(OnlineStateData.DATA,onlineState);
      content.put(OnlineStateData.LABEL,context.getString(R.string.label_onlineState));
      content.put(Data.MIMETYPE,OnlineStateData.CONTENT_ITEM_TYPE);
      addInsertOperation();
    }
    return this;
  }


  /***********************************************************************************************
  Adds a photo
  @param photoURL the URL to the photo
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addPhoto(URL photoURL){
    content.clear();

    if (photoURL != null && !photoURL.toString().equals("")) {
      // Download the photo into a byte[] BLOB (stored as a String)
      byte[] photoBytes = downloadPhoto(photoURL);

      // Only add a photo if the download was successful
      if (photoBytes != null && photoBytes.length > 0) {
        // store the date/time that this photo was retrieved, so its only grabbed once a day
        String timestamp = "" + Calendar.getInstance().getTimeInMillis();

        // Store the photo
        content.put(Photo.PHOTO,photoBytes);
        content.put(Photo.SYNC1,timestamp);
        content.put(Photo.MIMETYPE,Photo.CONTENT_ITEM_TYPE);
        addInsertOperation();
      }
    }
    return this;
  }


  /***********************************************************************************************
  Adds the primary group
  @param primaryGroup the primary group to add
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addPrimaryGroup(String primaryGroup){
    content.clear();
    if (primaryGroup != null && !primaryGroup.equals("")) {
      content.put(PrimaryGroupData.DATA,primaryGroup);
      content.put(PrimaryGroupData.LABEL,context.getString(R.string.label_primaryGroup));
      content.put(Data.MIMETYPE,PrimaryGroupData.CONTENT_ITEM_TYPE);
      addInsertOperation();
    }
    return this;
  }


  /***********************************************************************************************
  Adds the real name of a user
  @param realName the real name of the user
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addRealName(String realName){
    content.clear();
    if (realName != null && !realName.equals("")) {
      content.put(StructuredName.DISPLAY_NAME,realName);
      content.put(StructuredName.MIMETYPE,StructuredName.CONTENT_ITEM_TYPE);
    }
    if (content.size() > 0) {
      addInsertOperation();
    }
    return this;
  }


  /***********************************************************************************************
  Adds a "Steam" status message
  @param status the users status message
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addStatus(String status){

    content.put(StatusData.DATA,status);
    content.put(StatusData.LABEL,context.getString(R.string.label_status));
    content.put(Data.MIMETYPE,StatusData.CONTENT_ITEM_TYPE);
    addInsertOperation();

    return this;

  }


  /***********************************************************************************************
  Adds a "Steam" status icon
  @param status the users status message
  @param dataId the ID of the status data
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addStatusIcon(String status,long dataId){
    content.clear();

    if (dataId != -1) {

      content.put(ContactsContract.StatusUpdates.DATA_ID,dataId);
      content.put(ContactsContract.StatusUpdates.STATUS,status);
      content.put(ContactsContract.StatusUpdates.STATUS_RES_PACKAGE,GlobalConstants.ACCOUNT_TYPE);
      content.put(ContactsContract.StatusUpdates.STATUS_LABEL,context.getString(R.string.status_label));
      content.put(ContactsContract.StatusUpdates.STATUS_ICON,R.drawable.small_icon);
      content.put(ContactsContract.StatusUpdates.STATUS_TIMESTAMP,System.currentTimeMillis());
    }

    if (content.size() > 0) {
      // insert() does a replace() for status messages
      operationBuilder = addInsertOperation(addCallerIsSyncAdapterParameter(ContactsContract.StatusUpdates.CONTENT_URI),yield);
      operationBuilder.withValues(content);

      yield = false;
      batchOperation.add(operationBuilder.build());
      //batchOperation.execute();
    }

    return this;

  }


  /***********************************************************************************************
  Adds a steam rating
  @param steamRating the steam rating to add
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addSteamRating(String steamRating){
    content.clear();
    if (steamRating != null && !steamRating.equals("")) {
      content.put(SteamRatingData.DATA,steamRating);
      content.put(SteamRatingData.LABEL,context.getString(R.string.label_steamRating));
      content.put(Data.MIMETYPE,SteamRatingData.CONTENT_ITEM_TYPE);
      addInsertOperation();
    }
    return this;
  }


  /***********************************************************************************************
  Adds the summary
  @param summary the summary to add
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addSummary(String summary){
    content.clear();
    if (summary != null && !summary.equals("")) {
      content.put(SummaryData.DATA,summary);
      content.put(SummaryData.LABEL,context.getString(R.string.label_summary));
      content.put(Data.MIMETYPE,SummaryData.CONTENT_ITEM_TYPE);
      addInsertOperation();
    }
    return this;
  }


  /***********************************************************************************************
  Adds an update operation into the batch
  ***********************************************************************************************/
  public void addUpdateOperation(Uri uri){
    operationBuilder = addUpdateOperation(uri,yield).withValues(content);
    yield = false;
    batchOperation.add(operationBuilder.build());
  }


  /***********************************************************************************************
  Adds the users backend server userId
  @param userID the userID on the backend server
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addUserId(long userId){
    content.clear();
    if (userId >= 0) {
      content.put(UserIdData.DATA,userId);
      content.put(UserIdData.LABEL,context.getString(R.string.label_userId));
      content.put(Data.MIMETYPE,UserIdData.CONTENT_ITEM_TYPE);
      addInsertOperation();
    }
    return this;
  }


  /***********************************************************************************************
  Adds the username of a user
  @param username the username
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addUsername(String username){
    content.clear();
    if (username != null && !username.equals("")) {
      content.put(Nickname.NAME,username);
      content.put(Nickname.MIMETYPE,Nickname.CONTENT_ITEM_TYPE);
      content.put(Nickname.TYPE,Nickname.TYPE_CUSTOM);
      content.put(Nickname.LABEL,Language.get(R.string.label_nickname));
    }
    if (content.size() > 0) {
      addInsertOperation();
    }
    return this;
  }


  /***********************************************************************************************
  Adds a website
  @param website the website to add
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addWebsite(String website){
    content.clear();
    if (website != null && !website.equals("")) {
      content.put(WebsiteData.DATA,website);
      content.put(WebsiteData.LABEL,context.getString(R.string.label_website));
      content.put(Data.MIMETYPE,WebsiteData.CONTENT_ITEM_TYPE);
      addInsertOperation();
    }
    return this;
  }


  /***********************************************************************************************
  Adds the website title
  @param websiteTitle the website title to add
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations addWebsiteTitle(String websiteTitle){
    content.clear();
    if (websiteTitle != null && !websiteTitle.equals("")) {
      content.put(WebsiteTitleData.DATA,websiteTitle);
      content.put(WebsiteTitleData.LABEL,context.getString(R.string.label_websiteTitle));
      content.put(Data.MIMETYPE,WebsiteTitleData.CONTENT_ITEM_TYPE);
      addInsertOperation();
    }
    return this;
  }


  /***********************************************************************************************
  Builds a shoutout message based on the <i>shoutout</i> set by the friend and the <i>checkin</i>
  @param shoutout the users shoutout message
  @param checkin the last check-in location
  @return the shoutout message
  ***********************************************************************************************/
  public String buildShoutoutMessage(String shoutout,String checkin){
    if (shoutout == null || shoutout.equals("")) {
      if (checkin != null && !checkin.equals("")) {
        // Show the message "Checked in at <location>"
        shoutout = context.getString(R.string.checked_in_at) + " " + checkin;
      }
      else {
        // Set the initial shoutout to be empty, so we can still do status and icon updates, etc.
        shoutout = context.getString(R.string.no_shoutout_message);
      }
    }
    else if (checkin != null && !checkin.equals("")) {
      shoutout = "'" + shoutout + "' " + context.getString(R.string.while_watching) + " " + checkin;
    }

    return shoutout;
  }


  /***********************************************************************************************
  Updates the users photo
  @param photoURL the new photo URL
  @param contactUrl the Uri for the existing raw contact to be updated
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations deletePhoto(Uri contactUrl){
    addDeleteOperation(contactUrl);
    return this;
  }


  /***********************************************************************************************
  Downloads a photo from a <code>URL</code>. If the image is too large, it is resized.
  @param photoURL the URL to the photo
  @return the <code>byte</code> array of photo data (in JPEG format)
  ***********************************************************************************************/
  public byte[] downloadPhoto(URL photoURL){

    if (photoURL != null && !photoURL.toString().equals("")) {
      // Download the photo into a byte[] BLOB (stored as a String)
      byte[] photoBytes = new byte[1000];
      boolean photoSuccess = false;
      try {
        InputStream dataStream = photoURL.openStream();

        int readSize = 0;
        int receivedSize = dataStream.read(photoBytes,readSize,1000);

        while (receivedSize > 0) {
          // there is more data, so increase the array by another 1000 bytes and repeat the read
          readSize += receivedSize;
          photoBytes = ArrayResizer.resize(photoBytes,readSize + 1000);
          receivedSize = dataStream.read(photoBytes,readSize,1000);
        }

        dataStream.close();

        // resize to the actual size of the content
        photoBytes = ArrayResizer.resize(photoBytes,readSize);

        // Now check that the bytes are a valid JPEG file
        // (in case the URL was invalid or a 404 page was downloaded, etc)
        // (ie. only store the image if it is actually an image)
        Bitmap loadedJpg = BitmapFactory.decodeByteArray(photoBytes,0,readSize);
        if (loadedJpg != null) {
          // if the bitmap is larger than 150x150, resize it...
          int width = loadedJpg.getWidth();
          int height = loadedJpg.getHeight();
          if (width > 150 || height > 150) {
            width = 150;
            height = 150;
            // resize it
            loadedJpg = Bitmap.createScaledBitmap(loadedJpg,width,height,true);
            // get the new bytes for the resized image
            ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
            loadedJpg.compress(Bitmap.CompressFormat.JPEG,75,byteStream);
            photoBytes = byteStream.toByteArray();
          }
          photoSuccess = true;
        }
      }
      catch (Throwable t) {
        // Error downloading the photo
      }

      // Only add a photo if the download was successful
      if (photoSuccess) {
        return photoBytes;
      }

    }

    return null;
  }


  /***********************************************************************************************
  Formats a timestamp in a <code>Calendar</code> into a text <code>String</code>
  @param calendar the <code>Calendar</code> set to the timestamp
  @return the <code>String</code> representation of the timestamp
  ***********************************************************************************************/
  public String formatTimestamp(Calendar calendar){

    // Hour
    String timestamp = calendar.get(Calendar.HOUR) + ":";

    // Minute
    int minute = calendar.get(Calendar.MINUTE);
    if (minute < 10) {
      timestamp += "0";
    }
    timestamp += minute;

    // AM/PM
    if (calendar.get(Calendar.HOUR_OF_DAY) > 12) {
      timestamp += context.getString(R.string.time_pm) + " ";
    }
    else {
      timestamp += context.getString(R.string.time_am) + " ";
    }

    // Day
    int day = calendar.get(Calendar.DAY_OF_WEEK);
    if (day == Calendar.MONDAY) {
      timestamp += context.getString(R.string.day_monday) + " ";
    }
    else if (day == Calendar.TUESDAY) {
      timestamp += context.getString(R.string.day_tuesday) + " ";
    }
    else if (day == Calendar.WEDNESDAY) {
      timestamp += context.getString(R.string.day_wednesday) + " ";
    }
    else if (day == Calendar.THURSDAY) {
      timestamp += context.getString(R.string.day_thursday) + " ";
    }
    else if (day == Calendar.FRIDAY) {
      timestamp += context.getString(R.string.day_friday) + " ";
    }
    else if (day == Calendar.SATURDAY) {
      timestamp += context.getString(R.string.day_saturday) + " ";
    }
    else if (day == Calendar.SUNDAY) {
      timestamp += context.getString(R.string.day_sunday) + " ";
    }

    // Day number
    day = calendar.get(Calendar.DAY_OF_MONTH);
    timestamp += day;

    if (day == 1 || day == 21 || day == 31) {
      timestamp += context.getString(R.string.day_suffix_st) + " ";
    }
    else if (day == 2 || day == 22) {
      timestamp += context.getString(R.string.day_suffix_nd) + " ";
    }
    else if (day == 3 || day == 23) {
      timestamp += context.getString(R.string.day_suffix_rd) + " ";
    }
    else {
      timestamp += context.getString(R.string.day_suffix_th) + " ";
    }

    // Month
    int month = calendar.get(Calendar.MONTH);
    if (month == Calendar.JANUARY) {
      timestamp += context.getString(R.string.month_january) + " ";
    }
    else if (month == Calendar.FEBRUARY) {
      timestamp += context.getString(R.string.month_february) + " ";
    }
    else if (month == Calendar.MARCH) {
      timestamp += context.getString(R.string.month_march) + " ";
    }
    else if (month == Calendar.APRIL) {
      timestamp += context.getString(R.string.month_april) + " ";
    }
    else if (month == Calendar.MAY) {
      timestamp += context.getString(R.string.month_may) + " ";
    }
    else if (month == Calendar.JUNE) {
      timestamp += context.getString(R.string.month_june) + " ";
    }
    else if (month == Calendar.JULY) {
      timestamp += context.getString(R.string.month_july) + " ";
    }
    else if (month == Calendar.AUGUST) {
      timestamp += context.getString(R.string.month_august) + " ";
    }
    else if (month == Calendar.SEPTEMBER) {
      timestamp += context.getString(R.string.month_september) + " ";
    }
    else if (month == Calendar.OCTOBER) {
      timestamp += context.getString(R.string.month_october) + " ";
    }
    else if (month == Calendar.NOVEMBER) {
      timestamp += context.getString(R.string.month_november) + " ";
    }
    else if (month == Calendar.DECEMBER) {
      timestamp += context.getString(R.string.month_december) + " ";
    }

    // Year
    timestamp += calendar.get(Calendar.YEAR);

    return timestamp;
  }


  /***********************************************************************************************
  Updates a custom url
  @param customUrl the new custom url
  @param oldCustomUrl the old custom url
  @param contactUrl the Uri for the existing raw contact to be updated
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations updateCustomUrl(String customUrl,String oldCustomUrl,Uri contactUrl){
    if (customUrl != null && oldCustomUrl != null && !customUrl.equals(oldCustomUrl)) {
      content.clear();
      content.put(CustomUrlData.DATA,customUrl);
      addUpdateOperation(contactUrl);
    }
    return this;
  }


  /***********************************************************************************************
  Updates the favorite game
  @param favoriteGame the new favorite game
  @param oldFavoriteGame the old favorite game
  @param contactUrl the Uri for the existing raw contact to be updated
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations updateFavoriteGame(String favoriteGame,String oldFavoriteGame,Uri contactUrl){
    if (favoriteGame != null && oldFavoriteGame != null && !favoriteGame.equals(oldFavoriteGame)) {
      content.clear();
      content.put(FavoriteGameData.DATA,favoriteGame);
      addUpdateOperation(contactUrl);
    }
    return this;
  }


  /***********************************************************************************************
  Updates the headline
  @param headline the new headline
  @param oldHeadline the old headline
  @param contactUrl the Uri for the existing raw contact to be updated
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations updateHeadline(String headline,String oldHeadline,Uri contactUrl){
    if (headline != null && oldHeadline != null && !headline.equals(oldHeadline)) {
      content.clear();
      content.put(HeadlineData.DATA,headline);
      addUpdateOperation(contactUrl);
    }
    return this;
  }


  /***********************************************************************************************
  Updates the hours played
  @param hoursPlayed the new hours played
  @param oldHoursPlayed the old hours played
  @param contactUrl the Uri for the existing raw contact to be updated
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations updateHoursPlayed(String hoursPlayed,String oldHoursPlayed,Uri contactUrl){
    if (hoursPlayed != null && oldHoursPlayed != null && !hoursPlayed.equals(oldHoursPlayed)) {
      content.clear();
      content.put(HoursPlayedData.DATA,hoursPlayed);
      addUpdateOperation(contactUrl);
    }
    return this;
  }


  /***********************************************************************************************
  Updates the location
  @param location the new location
  @param oldLocation the old location
  @param contactUrl the Uri for the existing raw contact to be updated
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations updateLocation(String location,String oldLocation,Uri contactUrl){
    content.clear();
    if (location != null && oldLocation != null && !location.equals(oldLocation)) {
      content.put(StructuredPostal.FORMATTED_ADDRESS,location);

      String[] splitLocation = location.split(", ");
      int splitCount = splitLocation.length;
      if (splitCount == 1) {
        content.put(StructuredPostal.COUNTRY,splitLocation[0]);
      }
      else if (splitCount == 2) {
        content.put(StructuredPostal.REGION,splitLocation[0]);
        content.put(StructuredPostal.COUNTRY,splitLocation[1]);
      }
      else if (splitCount == 3) {
        content.put(StructuredPostal.CITY,splitLocation[0]);
        content.put(StructuredPostal.REGION,splitLocation[1]);
        content.put(StructuredPostal.COUNTRY,splitLocation[2]);
      }

    }

    if (content.size() > 0) {
      addUpdateOperation(contactUrl);
    }
    return this;
  }


  /***********************************************************************************************
  Updates a member since data
  @param memberSince the new member since data
  @param oldMemberSince the old member since data
  @param contactUrl the Uri for the existing raw contact to be updated
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations updateMemberSince(String memberSince,String oldMemberSince,Uri contactUrl){
    if (memberSince != null && oldMemberSince != null && !memberSince.equals(oldMemberSince)) {
      content.clear();
      content.put(MemberSinceData.DATA,memberSince);
      addUpdateOperation(contactUrl);
    }
    return this;
  }


  /***********************************************************************************************
  Updates the most played game
  @param mostPlayedGame the new most played game
  @param oldMostPlayedGame the old most played game
  @param contactUrl the Uri for the existing raw contact to be updated
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations updateMostPlayedGame(String mostPlayedGame,String oldMostPlayedGame,Uri contactUrl){
    if (mostPlayedGame != null && oldMostPlayedGame != null && !mostPlayedGame.equals(oldMostPlayedGame)) {
      content.clear();
      content.put(MostPlayedGameData.DATA,mostPlayedGame);
      addUpdateOperation(contactUrl);
    }
    return this;
  }


  /***********************************************************************************************
  Updates the most played game hours
  @param mostPlayedGameHours the new most played game hours
  @param oldMostPlayedGameHours the old most played game hours
  @param contactUrl the Uri for the existing raw contact to be updated
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations updateMostPlayedGameHours(String mostPlayedGameHours,String oldMostPlayedGameHours,Uri contactUrl){
    if (mostPlayedGameHours != null && oldMostPlayedGameHours != null && !mostPlayedGameHours.equals(oldMostPlayedGameHours)) {
      content.clear();
      content.put(MostPlayedGameHoursData.DATA,mostPlayedGameHours);
      addUpdateOperation(contactUrl);
    }
    return this;
  }


  /***********************************************************************************************
  Updates the online state
  @param onlineState the new online state
  @param oldOnlineState the old online state
  @param contactUrl the Uri for the existing raw contact to be updated
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations updateOnlineState(String onlineState,String oldOnlineState,Uri contactUrl){
    if (onlineState != null && oldOnlineState != null && !onlineState.equals(oldOnlineState)) {
      content.clear();
      content.put(OnlineStateData.DATA,onlineState);
      addUpdateOperation(contactUrl);
    }
    return this;
  }


  /***********************************************************************************************
  Updates the users photo
  @param photoURL the new photo URL
  @param contactUrl the Uri for the existing raw contact to be updated
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations updatePhoto(URL photoURL,Uri contactUrl){
    if (photoURL != null && !photoURL.toString().equals("")) {
      content.clear();

      // Download the photo into a byte[] BLOB (stored as a String)
      byte[] photoBytes = downloadPhoto(photoURL);

      // Only update the photo if the download was successful
      if (photoBytes != null && photoBytes.length > 0) {
        // Store the photo
        content.put(Photo.PHOTO,photoBytes);
        addUpdateOperation(contactUrl);
      }

    }
    return this;
  }


  /***********************************************************************************************
  Updates the primary group
  @param primaryGroup the new primary group
  @param oldPrimaryGroup the old primary group
  @param contactUrl the Uri for the existing raw contact to be updated
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations updatePrimaryGroup(String primaryGroup,String oldPrimaryGroup,Uri contactUrl){
    if (primaryGroup != null && oldPrimaryGroup != null && !primaryGroup.equals(oldPrimaryGroup)) {
      content.clear();
      content.put(PrimaryGroupData.DATA,primaryGroup);
      addUpdateOperation(contactUrl);
    }
    return this;
  }


  /***********************************************************************************************
  Updates the users real name
  @param realName the new name
  @param oldRealName the old name
  @param contactUrl the Uri for the existing raw contact to be updated
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations updateRealName(String realName,String oldRealName,Uri contactUrl){
    content.clear();
    if (realName != null && oldRealName != null && !realName.equals(oldRealName)) {
      content.put(StructuredName.DISPLAY_NAME,realName);
    }
    if (content.size() > 0) {
      addUpdateOperation(contactUrl);
    }
    return this;
  }


  /***********************************************************************************************
  Updates a "Steam" status message
  @param shoutout the users shoutout message
  @param oldShoutout the old shoutout
  @param checkin the last check-in location
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations updateStatus(String status,String oldStatus,Uri contactUrl){
    if (status == null || oldStatus == null || (status != null && oldStatus != null && !status.equals(oldStatus))) {
      content.put(StatusData.DATA,status);
      addUpdateOperation(contactUrl);
    }

    return this;

  }


  /***********************************************************************************************
  Updates the "Steam" status icon
  @param shoutout the new status
  @param oldShoutout the old status
  @param dataId the ID of the status data
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations updateStatusIcon(String status,String oldStatus,long dataId){

    if (status == null || oldStatus == null || (status != null && oldStatus != null && !status.equals(oldStatus))) {
      return addStatusIcon(status,dataId);
    }
    return this;
  }


  /***********************************************************************************************
  Updates a steam rating
  @param steamRating the new steam rating
  @param oldSteamRating the old steam rating
  @param contactUrl the Uri for the existing raw contact to be updated
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations updateSteamRating(String steamRating,String oldSteamRating,Uri contactUrl){
    if (steamRating != null && oldSteamRating != null && !steamRating.equals(oldSteamRating)) {
      content.clear();
      content.put(SteamRatingData.DATA,steamRating);
      addUpdateOperation(contactUrl);
    }
    return this;
  }


  /***********************************************************************************************
  Updates the summary
  @param summary the new summary
  @param oldSummary the old summary
  @param contactUrl the Uri for the existing raw contact to be updated
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations updateSummary(String summary,String oldSummary,Uri contactUrl){
    if (summary != null && oldSummary != null && !summary.equals(oldSummary)) {
      content.clear();
      content.put(SummaryData.DATA,summary);
      addUpdateOperation(contactUrl);
    }
    return this;
  }


  /***********************************************************************************************
  Updates the users backend server user ID
  @param userId the new user ID
  @param oldUserId the old user ID
  @param contactUrl the Uri for the existing raw contact to be updated
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations updateUserId(long userId,long oldUserId,Uri contactUrl){
    if (userId >= 0 && oldUserId >= 0 && userId != oldUserId) {
      content.clear();
      content.put(UserIdData.DATA,userId);
      addUpdateOperation(contactUrl);
    }
    return this;
  }


  /***********************************************************************************************
  Updates the users username
  @param username the new username
  @param oldUsername the old username
  @param contactUrl the Uri for the existing raw contact to be updated
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations updateUsername(String username,String oldUsername,Uri contactUrl){
    content.clear();
    if (username != null && oldUsername != null && !username.equals(oldUsername)) {
      content.put(Nickname.NAME,username);
    }
    if (content.size() > 0) {
      addUpdateOperation(contactUrl);
    }
    return this;
  }


  /***********************************************************************************************
  Updates the website
  @param website the new website
  @param oldWebsite the old website
  @param contactUrl the Uri for the existing raw contact to be updated
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations updateWebsite(String website,String oldWebsite,Uri contactUrl){
    if (website != null && oldWebsite != null && !website.equals(oldWebsite)) {
      content.clear();
      content.put(WebsiteData.DATA,website);
      addUpdateOperation(contactUrl);
    }
    return this;
  }


  /***********************************************************************************************
  Updates the website title
  @param mostPlayedGameHours the new website title
  @param oldMostPlayedGameHours the old website title
  @param contactUrl the Uri for the existing raw contact to be updated
  @return instance of ContactOperations
  ***********************************************************************************************/
  public ContactOperations updateWebsiteTitle(String websiteTitle,String oldWebsiteTitle,Uri contactUrl){
    if (websiteTitle != null && oldWebsiteTitle != null && !websiteTitle.equals(oldWebsiteTitle)) {
      content.clear();
      content.put(WebsiteTitleData.DATA,websiteTitle);
      addUpdateOperation(contactUrl);
    }
    return this;
  }


  /***********************************************************************************************
  Adds a CALLER_IS_SYNCADAPTER parameter
  ***********************************************************************************************/
  public static Uri addCallerIsSyncAdapterParameter(Uri uri){
    return uri.buildUpon().appendQueryParameter(ContactsContract.CALLER_IS_SYNCADAPTER,"true").build();
  }


  /***********************************************************************************************
  Builds a <code>ContentProviderOperation</code> for a delete operation
  ***********************************************************************************************/
  public static ContentProviderOperation.Builder addDeleteOperation(Uri uri,boolean yield){
    return ContentProviderOperation.newDelete(addCallerIsSyncAdapterParameter(uri)).withYieldAllowed(yield);

  }


  /***********************************************************************************************
  Builds a <code>ContentProviderOperation</code> for an insert operation
   ***********************************************************************************************/
  public static ContentProviderOperation.Builder addInsertOperation(Uri uri,boolean yield){
    return ContentProviderOperation.newInsert(addCallerIsSyncAdapterParameter(uri)).withYieldAllowed(yield);
  }


  /***********************************************************************************************
  Builds a <code>ContentProviderOperation</code> for an update operation
   ***********************************************************************************************/
  public static ContentProviderOperation.Builder addUpdateOperation(Uri uri,boolean yield){
    return ContentProviderOperation.newUpdate(addCallerIsSyncAdapterParameter(uri)).withYieldAllowed(yield);
  }


  /***********************************************************************************************
  Creates a new <code>ContactOperations</code> for adding new contact to the Android Contacts
  @param context the Authenticator Activity context
  @param userId the userId of the sample SyncAdapter user object
  @param accountName the username of the current login
  @param batchOperations any batch operations to perform when creating the new contact 
  @return instance of ContactOperations
  ***********************************************************************************************/
  public static ContactOperations createNewContact(Context context,int userId,String accountName,BatchOperation batchOperations){
    return new ContactOperations(context,userId,accountName,batchOperations);
  }


  /***********************************************************************************************
  Creates a new <code>ContactOperations</code> for adding new group to the Android Contacts
  @param context the Authenticator Activity context
  @param groupId the groupId of the sample SyncAdapter user object
  @param accountName the username of the current login
  @param batchOperations any batch operations to perform when creating the new contact 
  @return instance of ContactOperations
  ***********************************************************************************************/
  public static ContactOperations createNewGroup(Context context,int groupId,String accountName,BatchOperation batchOperations){
    return new ContactOperations(context,groupId,accountName,batchOperations,true);
  }


  /***********************************************************************************************
  Creates a new <code>ContactOperations</code> for updating an existing Android Contact
  @param context the Authenticator Activity context
  @param rawContactId the unique ID of the existing contact
  @param batchOperations any batch operations to perform when updating the contact 
  @return instance of ContactOperations
   ***********************************************************************************************/
  public static ContactOperations updateExistingContact(Context context,long rawContactId,BatchOperation batchOperations){
    return new ContactOperations(context,rawContactId,batchOperations);
  }

}
