////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                            //
//                                       WATTO STUDIOS                                        //
//                             Java Code, Programs, and Software                              //
//                                    http://www.watto.org                                    //
//                                                                                            //
//                           Copyright (C) 2004-2010  WATTO Studios                           //
//                                                                                            //
// This program is free software; you can redistribute it and/or modify it under the terms of //
// the GNU General Public License published by the Free Software Foundation; either version 2 //
// of the License, or (at your option) any later versions. This program is distributed in the //
// hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranties //
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License //
// at http://www.gnu.org for more details. For updates and information about this program, go //
// to the WATTO Studios website at http://www.watto.org or email watto@watto.org . Thanks! :) //
//                                                                                            //
////////////////////////////////////////////////////////////////////////////////////////////////

package org.watto.program.android.sync.steam;

import java.net.URL;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.util.EntityUtils;
import org.watto.android.xml.XMLNode;
import org.watto.android.xml.XMLReader;
import org.watto.program.android.sync.steam.authenticator.SteamAuthenticator;


/***********************************************************************************************
 Provides utility methods for communicating with the backend server
***********************************************************************************************/
public class SteamQuery {

  /***********************************************************************************************
  Fetches the list of friend data updates from the server
  @return list The list of updates received from the server.
  ***********************************************************************************************/
  public static SteamFriend[] getFriends(){
    try {

      if (!SteamAuthenticator.credentialsKnown()) {
        // The authentication is invalid, so return nothing
        return new SteamFriend[0];
      }

      XMLNode profileNode = queryServer(SteamAuthenticator.getUserUrl());
      if (profileNode != null && profileNode.hasChildren()) {

        // Get the friends branch
        XMLNode friendsNode = profileNode.getChild("friends");
        if (friendsNode != null && friendsNode.hasChildren()) {
          // Yay, we have friends!
          int friendCount = friendsNode.getChildCount();

          SteamFriend[] friends = new SteamFriend[friendCount];

          // Create the friends
          for (int i = 0;i < friendCount;i++) {
            XMLNode friendNode = friendsNode.getChild(i);
            if (!friendNode.getTag().equals("friend")) {
              // Not a "friend" node for some reason - bad XML!
              continue;
            }

            long userID = -1;
            try {
              userID = Long.parseLong(friendNode.getChild("steamID64").getContent());
            }
            catch (Throwable t) {
              continue;
            }

            friends[i] = new SteamFriend(userID);
          }

          // get further details about each friend
          int validFriendCount = getFriendsDetails(friends);
          
          if (validFriendCount != friendCount){
            // At least 1 friend did not have a public profile, so want to shrink it down
            
            SteamFriend[] oldFriends = friends;
            friends = new SteamFriend[validFriendCount];
            
            for (int f=0,v=0;f<friendCount && v<validFriendCount;f++){
              SteamFriend friend = oldFriends[f];
              if (friend != null){
                friends[v] = friend;
                v++;
              }
            }
            
            
          }

          return friends;
        }

      }

      return new SteamFriend[0];
    }
    catch (Throwable t) {
      // An error occurred in the response, or there were parsing errors
      return new SteamFriend[0];
    }
  }


  /***********************************************************************************************
  Gets further details about each friend
  @param friends the friends to get the details for
  @return the number of friends that had public profiles
  ***********************************************************************************************/
  public static int getFriendsDetails(SteamFriend[] friends){
    try {

      int friendCount = friends.length;
      
      // Used to count up the number of friends that have public profiles
      int validFriendCount = friendCount;
      
      for (int i = 0;i < friendCount;i++) {
        SteamFriend friend = friends[i];

        long userID = friend.getUserID();

        XMLNode profileNode = queryServer(SteamAuthenticator.getUserUrl(userID));
        if (profileNode != null && profileNode.hasChildren()) {
          
          try {
            friend.setUsername(profileNode.getChild("steamID").getContent());
          }
          catch (Throwable t) {}

          try {
            friend.setOnlineState(profileNode.getChild("onlineState").getContent());
          }
          catch (Throwable t) {}

          try {
            friend.setStateMessage(profileNode.getChild("stateMessage").getContent().replaceAll("<br />",". "));
          }
          catch (Throwable t) {}

          try {
            friend.setPhoto(new URL(profileNode.getChild("avatarFull").getContent()));
          }
          catch (Throwable t) {}

          try {
            friend.setCustomUrl(profileNode.getChild("customURL").getContent());
          }
          catch (Throwable t) {}

          try {
            friend.setMemberSince(profileNode.getChild("memberSince").getContent());
          }
          catch (Throwable t) {}

          try {
            friend.setSteamRating(Double.parseDouble(profileNode.getChild("steamRating").getContent()));
          }
          catch (Throwable t) {}

          try {
            friend.setHoursPlayed(Double.parseDouble(profileNode.getChild("hoursPlayed2Wk").getContent()));
          }
          catch (Throwable t) {}

          try {
            friend.setHeadline(profileNode.getChild("headline").getContent().replaceAll("<br />",". "));
          }
          catch (Throwable t) {}

          try {
            friend.setLocation(profileNode.getChild("location").getContent());
          }
          catch (Throwable t) {}

          try {
            friend.setRealName(profileNode.getChild("realname").getContent());
          }
          catch (Throwable t) {}

          try {
            friend.setSummary(profileNode.getChild("summary").getContent().replaceAll("<br />",". "));
          }
          catch (Throwable t) {}

          // Most Played Game
          try {
            XMLNode mostPlayedGameNode = profileNode.getChild("mostPlayedGames").getChild("mostPlayedGame");

            friend.setMostPlayedGame(mostPlayedGameNode.getChild("gameName").getContent());
            friend.setMostPlayedGameHours(Double.parseDouble(mostPlayedGameNode.getChild("hoursOnRecord").getContent()));
          }
          catch (Throwable t) {}

          // Website
          try {
            XMLNode weblinkNode = profileNode.getChild("weblinks").getChild("weblink");

            friend.setWebsite(weblinkNode.getChild("link").getContent());
            friend.setWebsiteTitle(weblinkNode.getChild("title").getContent().replaceAll("<br />",". "));
          }
          catch (Throwable t) {}

          // Primary Group
          try {
            XMLNode groupsNode = profileNode.getChild("groups");
            int groupCount = groupsNode.getChildCount();

            for (int g = 0;g < groupCount;g++) {
              XMLNode groupNode = groupsNode.getChild(g);
              if (!groupNode.getTag().equals("group")) {
                continue;
              }
              if (!groupNode.getAttribute("isPrimary").equals("1")) {
                continue;
              }

              friend.setPrimaryGroup(groupNode.getChild("groupName").getContent().replaceAll("<br />",". "));
              // if we're here, we have found the primary group;
              break;
            }
          }
          catch (Throwable t) {}

          // Favorite Game
          try {
            XMLNode favoriteGameNode = profileNode.getChild("favoriteGame");

            friend.setFavoriteGame(favoriteGameNode.getChild("name").getContent());
          }
          catch (Throwable t) {}
          
          
          // See if there is a privacy message - if so, this user's profile is private or not created
          try {
            XMLNode privacyNode = profileNode.getChild("privacyMessage");
            String privacyContent = privacyNode.getContent();
            if (privacyContent != null && !privacyContent.equals("null") && !privacyContent.equals("")){
              // There is some kind of privacy content here, so don't show this user
              friends[i] = null;
              validFriendCount--;
            }
          }
          catch (Throwable t) {}


        }

      }

      return validFriendCount;
    }
    catch (Throwable t) {
      // An error occurred in the response, or there were parsing errors
      return 0;
    }
  }


  /***********************************************************************************************
  Signs a query, sends it to the server, receives a response and returns it 
  @param queryUrl the URL of the query you wish to perform
  @return the response as an <code>XMLNode</code>, or <b>null</b> if there was an error
  ***********************************************************************************************/
  public static XMLNode queryServer(String queryUrl){
    try {

      HttpGet request = new HttpGet(queryUrl);

      // Send the request
      DefaultHttpClient client = new DefaultHttpClient();
      HttpResponse response = client.execute(request);

      // Get the response code
      if (response.getStatusLine().getStatusCode() == HttpStatus.SC_OK) {
        // Success
        String responseMessage = EntityUtils.toString(response.getEntity());

        // Convert the response into an XML tree
        XMLNode xmlTree = XMLReader.read(responseMessage);

        // return the xml response
        return xmlTree;
      }
      return null;
    }
    catch (Throwable t) {
      return null;
    }
  }



}
