////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                            //
//                                       WATTO STUDIOS                                        //
//                             Java Code, Programs, and Software                              //
//                                    http://www.watto.org                                    //
//                                                                                            //
//                           Copyright (C) 2004-2010  WATTO Studios                           //
//                                                                                            //
// This program is free software; you can redistribute it and/or modify it under the terms of //
// the GNU General Public License published by the Free Software Foundation; either version 2 //
// of the License, or (at your option) any later versions. This program is distributed in the //
// hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranties //
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License //
// at http://www.gnu.org for more details. For updates and information about this program, go //
// to the WATTO Studios website at http://www.watto.org or email watto@watto.org . Thanks! :) //
//                                                                                            //
////////////////////////////////////////////////////////////////////////////////////////////////

package org.watto.program.android.sync.steam.sync;

import java.util.Calendar;
import org.watto.android.R;
import org.watto.program.android.sync.steam.BatchOperation;
import org.watto.program.android.sync.steam.ContactOperations;
import org.watto.program.android.sync.steam.GlobalConstants;
import org.watto.program.android.sync.steam.SteamFriend;
import org.watto.program.android.sync.steam.SteamQuery;
import org.watto.program.android.sync.steam.datatype.CustomUrlData;
import org.watto.program.android.sync.steam.datatype.FavoriteGameData;
import org.watto.program.android.sync.steam.datatype.HeadlineData;
import org.watto.program.android.sync.steam.datatype.HoursPlayedData;
import org.watto.program.android.sync.steam.datatype.MemberSinceData;
import org.watto.program.android.sync.steam.datatype.MostPlayedGameData;
import org.watto.program.android.sync.steam.datatype.MostPlayedGameHoursData;
import org.watto.program.android.sync.steam.datatype.OnlineStateData;
import org.watto.program.android.sync.steam.datatype.PrimaryGroupData;
import org.watto.program.android.sync.steam.datatype.StatusData;
import org.watto.program.android.sync.steam.datatype.SteamRatingData;
import org.watto.program.android.sync.steam.datatype.SummaryData;
import org.watto.program.android.sync.steam.datatype.UserIdData;
import org.watto.program.android.sync.steam.datatype.WebsiteData;
import org.watto.program.android.sync.steam.datatype.WebsiteTitleData;
import android.accounts.Account;
import android.accounts.AccountManager;
import android.content.AbstractThreadedSyncAdapter;
import android.content.ContentProviderClient;
import android.content.ContentResolver;
import android.content.ContentUris;
import android.content.Context;
import android.content.SyncResult;
import android.database.Cursor;
import android.net.Uri;
import android.os.Bundle;
import android.provider.ContactsContract;
import android.provider.ContactsContract.CommonDataKinds.GroupMembership;
import android.provider.ContactsContract.CommonDataKinds.Nickname;
import android.provider.ContactsContract.CommonDataKinds.Photo;
import android.provider.ContactsContract.CommonDataKinds.StructuredName;
import android.provider.ContactsContract.CommonDataKinds.StructuredPostal;
import android.provider.ContactsContract.Data;
import android.provider.ContactsContract.Groups;
import android.provider.ContactsContract.RawContacts;
import android.provider.ContactsContract.RawContacts.Entity;


/***********************************************************************************************
A <code>SyncAdapter</code> implementation for syncing <code>SyncAdapter</code> contacts to the
platform <code>ContactOperations</code> provider.
***********************************************************************************************/
public class SyncAdapter extends AbstractThreadedSyncAdapter {

  /** Access to the sync accounts on this device **/
  AccountManager accountManager;
  Context context;


  /***********************************************************************************************
  Constructor
  ***********************************************************************************************/
  public SyncAdapter(Context context, boolean autoInitialize){
    super(context,autoInitialize);
    this.context = context;
    this.accountManager = AccountManager.get(context);
  }


  /***********************************************************************************************
  Performs a synchronisation for this account. SyncAdapter-specific parameters may be specified in
  extras, which is guaranteed to not be null.
  @param account the account that should be synced 
  @param extras SyncAdapter-specific parameters 
  @param authority the authority of this sync request 
  @param provider a ContentProviderClient that points to the ContentProvider for this authority 
  @param syncResult SyncAdapter-specific parameters  
  ***********************************************************************************************/
  public void onPerformSync(Account account,Bundle extras,String authority,ContentProviderClient provider,SyncResult syncResult){
    try {
      // Get the list of friends, and sync them with the Android Contacts
      SteamFriend[] friends = SteamQuery.getFriends();
      syncContacts(context,account.name,friends);
    }
    catch (Throwable t) {
      syncResult.stats.numIoExceptions++;
    }
  }


  /***********************************************************************************************
  Adds a single contact to the platform contacts provider
  @param context the Authenticator Activity context
  @param accountName the account the contact belongs to
  @param friend the user to add
  @param rawGroupId the unique Id for the "Steam" group
  @param rawStatusId the unique Id for the "Steam" status message
  @param batchOperations any batch operations to perform when creating the new contact 
  ***********************************************************************************************/
  public static void addContact(Context context,String accountName,SteamFriend friend,long rawGroupId,long rawStatusId,BatchOperation batchOperations){
    // Creates a new ContactOperation batch process
    int userID = -1; // the raw userID will be populated when the contact is created
    ContactOperations contactOperation = ContactOperations.createNewContact(context,userID,accountName,batchOperations);

    // Adds the content to the operation
    contactOperation.addUserId(friend.getUserID());
    contactOperation.addUsername(friend.getUsername());
    contactOperation.addOnlineState(friend.getOnlineState());
    contactOperation.addPhoto(friend.getPhoto());
    contactOperation.addCustomUrl(friend.getCustomUrl());
    contactOperation.addMemberSince(friend.getMemberSince());
    contactOperation.addSteamRating("" + friend.getSteamRating());
    contactOperation.addHoursPlayed("" + friend.getHoursPlayed());
    contactOperation.addHeadline(friend.getHeadline());
    contactOperation.addLocation(friend.getLocation());
    contactOperation.addRealName(friend.getRealName());
    contactOperation.addSummary(friend.getSummary());
    contactOperation.addMostPlayedGame(friend.getMostPlayedGame());
    contactOperation.addMostPlayedGameHours("" + friend.getMostPlayedGameHours());
    contactOperation.addWebsite(friend.getWebsite());
    contactOperation.addWebsiteTitle(friend.getWebsiteTitle());
    contactOperation.addPrimaryGroup(friend.getPrimaryGroup());
    contactOperation.addFavoriteGame(friend.getFavoriteGame());

    contactOperation.addStatus(friend.getStateMessage());

    // Note: Don't add the status icon here, as the data for this user doesn't exist yet.
    // It is done in the syncContacts() instead.
    //contactOperation.addStatusIcon(friend.getShoutout(),friend.getCheckin(),rawStatusId);

    contactOperation.addGroupMembership(rawGroupId);
  }


  /***********************************************************************************************
  Synchronise raw contacts
  @param context The context of Authenticator Activity
  @param accountName The username for the account
  @param users The list of users
  ***********************************************************************************************/
  public static void addGroup(Context context,String accountName,BatchOperation batchOperation){

    // Creates a new ContactOperation batch process
    int groupID = -1; // the raw groupID will be populated when the group is created
    ContactOperations contactOperation = ContactOperations.createNewGroup(context,groupID,accountName,batchOperation);

    String title = context.getString(R.string.contact_group_title);
    String notes = context.getString(R.string.contact_group_notes);
    contactOperation.addGroup(title,notes);
  }


  /***********************************************************************************************
  Deletes a contact from the platform contacts provider
  @param context the Authenticator Activity context
  @param rawContactId the unique Id for this rawContact in contacts provider
  @param batchOperations any batch operations to perform when deleting the contact 
  ***********************************************************************************************/
  public static void deleteContact(Context context,long rawContactId,BatchOperation batchOperation){
    batchOperation.add(ContactOperations.addDeleteOperation(ContentUris.withAppendedId(RawContacts.CONTENT_URI,rawContactId),true).build());
  }


  /***********************************************************************************************
  Gets the platforms' rawContactID for the Steam group
  @param resolver the ContentResolver for querying the platform contacts
  @return the rawGroupID for the group, or <b>0</b> if no matching group was found
  ***********************************************************************************************/
  public static long getGroupID(ContentResolver resolver){

    String[] columns = new String[]{Groups._ID};
    String whereClause = ContactsContract.Groups.ACCOUNT_TYPE + "='" + GlobalConstants.ACCOUNT_TYPE + "'" + " AND " + ContactsContract.Groups.ACCOUNT_NAME + "='" + GlobalConstants.ACCOUNT_NAME + "'";

    Cursor c = resolver.query(ContactsContract.Groups.CONTENT_URI,columns,whereClause,null,null);

    long rawGroupID = -1;
    try {
      if (c != null && c.moveToFirst()) {
        rawGroupID = c.getLong(0);
      }
    }
    finally {
      if (c != null) {
        c.close();
      }
    }
    return rawGroupID;
  }


  /***********************************************************************************************
  Gets the platforms' rawContactID for the contact <code>SteamFriend.getUserID()</code>
  @param resolver the ContentResolver for querying the platform contacts
  @param userID the <code>SteamFriend.getUserID()</code> to lookup
  @return the rawContactID for the contact, or <b>0</b> if no matching contact was found
  ***********************************************************************************************/
  public static long getRawContactID(ContentResolver resolver,long userID){

    //String[] columns = new String[]{RawContacts._ID};
    String[] columns = new String[]{Data.RAW_CONTACT_ID};
    String whereClause = RawContacts.ACCOUNT_TYPE + "='" + GlobalConstants.ACCOUNT_TYPE + "' AND " + Data.MIMETYPE + "='" + UserIdData.CONTENT_ITEM_TYPE + "' AND " + UserIdData.DATA + "=" + userID;

    long rawContactID = 0;
    Cursor c = resolver.query(Data.CONTENT_URI,columns,whereClause,null,null);
    try {
      if (c != null && c.moveToFirst()) {
        rawContactID = c.getLong(0);
      }
    }
    finally {
      if (c != null) {
        c.close();
      }
    }
    return rawContactID;
  }


  /***********************************************************************************************
  Gets the platforms' rawContactID for the Steam status
  @param resolver the ContentResolver for querying the platform contacts
  @return the rawContactID for the contact, or <b>0</b> if no matching contact was found
  ***********************************************************************************************/
  public static long getStatusDataID(ContentResolver resolver,long rawContactId){
    if (rawContactId == -1) {
      return -1;
    }

    Uri rawContactUri = ContentUris.withAppendedId(RawContacts.CONTENT_URI,rawContactId);
    Uri entityUri = Uri.withAppendedPath(rawContactUri,Entity.CONTENT_DIRECTORY);

    String[] columns = new String[]{Entity.DATA_ID,Entity.MIMETYPE};
    String whereClause = Entity.MIMETYPE + "='" + StatusData.CONTENT_ITEM_TYPE + "'";

    long rawStatusId = -1;
    try {
      //Cursor c = resolver.query(entityUri,columns,null,null,null);
      Cursor c = resolver.query(entityUri,columns,whereClause,null,null);
      try {
        while (c != null && c.moveToNext()) {
          String mimetype = c.getString(1);
          // Store the Status against the Status data of the user
          if (mimetype.equals(StatusData.CONTENT_ITEM_TYPE)) {
            rawStatusId = c.getLong(0);
            return rawStatusId;
          }

        }
      }
      finally {
        if (c != null) {
          c.close();
        }
      }
    }
    catch (Throwable t) {}
    return rawStatusId;
  }


  /***********************************************************************************************
  Synchronise raw contacts
  @param context The context of Authenticator Activity
  @param accountName The name of the account
  @param users The list of users
  ***********************************************************************************************/
  public static synchronized void syncContacts(Context context,String accountName,SteamFriend[] friends){
    if (friends == null || friends.length <= 0) {
      return;
    }

    long userID;
    long rawContactId = 0;
    long rawGroupId = -1;
    long rawStatusId = -1;

    ContentResolver resolver = context.getContentResolver();
    BatchOperation batchOperation = new BatchOperation(context,resolver);

    rawGroupId = getGroupID(resolver);
    if (rawGroupId == -1) {
      // Create the group
      addGroup(context,accountName,batchOperation);
      // Get the group ID
      rawGroupId = getGroupID(resolver);
    }

    int friendCount = friends.length;
    for (int i = 0;i < friendCount;i++) {
      SteamFriend friend = friends[i];

      userID = friend.getUserID();
      // Check to see if the contact needs to be inserted or updated
      rawContactId = getRawContactID(resolver,userID);

      // Update the users Status icon and message
      rawStatusId = getStatusDataID(resolver,rawContactId);

      if (rawContactId != 0) {
        // update contact
        updateContact(context,resolver,accountName,friend,rawContactId,rawGroupId,rawStatusId,batchOperation);
      }
      else {
        // add new contact
        addContact(context,accountName,friend,rawGroupId,rawStatusId,batchOperation);
        // create the contact now, so we can update the status
        batchOperation.execute();
        // now get the rawContactID and statusDataID, then update the status
        rawContactId = getRawContactID(resolver,userID);
        rawStatusId = getStatusDataID(resolver,rawContactId);
        new ContactOperations(context,batchOperation).addStatusIcon(friend.getStateMessage(),rawStatusId);
      }

      // Execute and clear the batch operations if it gets too large
      if (batchOperation.size() >= 50) {
        batchOperation.execute();
      }
    }

    // Execute the batch operations
    batchOperation.execute();
  }


  /***********************************************************************************************
  Updates a single contact to the platform contacts provider.
  @param context the Authenticator Activity context
  @param resolver the ContentResolver to use
  @param accountName the account the contact belongs to
  @param friend the user being updated
  @param rawContactId the unique Id for this rawContact in contacts provider
  @param rawGroupId the unique Id for the "Steam" group
  @param rawStatusId the unique Id for the "Steam" status message
  ***********************************************************************************************/
  public static void updateContact(Context context,ContentResolver resolver,String accountName,SteamFriend friend,long rawContactId,long rawGroupId,long rawStatusId,BatchOperation batchOperation){
    Uri uri;

    long oldUserID = -1;
    String oldUsername = null;
    String oldOnlineState = null;
    String oldStatus = null;
    //URL oldPhoto = null; // SPECIAL CASE - See lower down!
    boolean photoUpdated = false;
    String oldCustomUrl = null;
    String oldMemberSince = null;
    String oldSteamRating = null;
    String oldHoursPlayed = null;
    String oldHeadline = null;
    String oldLocation = null;
    String oldRealName = null;
    String oldSummary = null;
    String oldMostPlayedGame = null;
    String oldMostPlayedGameHours = null;
    String oldWebsite = null;
    String oldWebsiteTitle = null;
    String oldPrimaryGroup = null;
    String oldFavoriteGame = null;

    long oldGroupMembership = -1;

    ContactOperations contactOperation = ContactOperations.updateExistingContact(context,rawContactId,batchOperation);

    // BUG FIX!
    // There was a bug in 1.0 and 1.1 where a new photo would be added to the contact every day!
    // This code here is to detect and remove all the photo details for the contact, if there are
    // more than 1, so that it can update the existing photo rather than adding new ones!
    String[] photoBugColumns = new String[]{Data._ID};
    String photoBugWhereClause = Data.RAW_CONTACT_ID + "=" + rawContactId + " AND " + Data.MIMETYPE + "='" + Photo.CONTENT_ITEM_TYPE + "'";
    Cursor photoBugC = resolver.query(Data.CONTENT_URI,photoBugColumns,photoBugWhereClause,null,null);
    int count = photoBugC.getCount();
    if (count > 1) {
      // remove all the photo entries
      while (photoBugC.moveToNext()) {
        long columnId = photoBugC.getLong(0);
        uri = ContentUris.withAppendedId(Data.CONTENT_URI,columnId);
        contactOperation.deletePhoto(uri);
      }
      // Clear the photos right now, so we can get the correct Photos in the next Cursor SELECT
      batchOperation.execute();
    }
    photoBugC.close();

    // Now the real update query...
    String[] columns = new String[]{Data._ID,Data.MIMETYPE,Data.DATA1,Data.DATA2,Data.DATA3,Data.DATA4,Data.DATA5,Data.DATA6,Data.DATA7,Data.DATA8,Data.DATA9,Data.DATA10,Data.SYNC1};
    String whereClause = Data.RAW_CONTACT_ID + "=" + rawContactId;
    Cursor c = resolver.query(Data.CONTENT_URI,columns,whereClause,null,null);

    try {
      while (c.moveToNext()) {
        long columnId = c.getLong(0);
        uri = ContentUris.withAppendedId(Data.CONTENT_URI,columnId);

        String mimeType = c.getString(1);

        if (mimeType.equals(UserIdData.CONTENT_ITEM_TYPE)) {
          oldUserID = c.getLong(2);
          contactOperation.updateUserId(friend.getUserID(),oldUserID,uri);
        }
        else if (mimeType.equals(Nickname.CONTENT_ITEM_TYPE)) {
          oldUsername = c.getString(2);
          contactOperation.updateUsername(friend.getUsername(),oldUsername,uri);
        }
        else if (mimeType.equals(OnlineStateData.CONTENT_ITEM_TYPE)) {
          oldOnlineState = c.getString(2);
          contactOperation.updateOnlineState(friend.getOnlineState(),oldOnlineState,uri);
        }
        else if (mimeType.equals(CustomUrlData.CONTENT_ITEM_TYPE)) {
          oldCustomUrl = c.getString(2);
          contactOperation.updateCustomUrl(friend.getCustomUrl(),oldCustomUrl,uri);
        }
        else if (mimeType.equals(MemberSinceData.CONTENT_ITEM_TYPE)) {
          oldMemberSince = c.getString(2);
          contactOperation.updateMemberSince(friend.getMemberSince(),oldMemberSince,uri);
        }
        else if (mimeType.equals(SteamRatingData.CONTENT_ITEM_TYPE)) {
          oldSteamRating = c.getString(2);
          contactOperation.updateSteamRating("" + friend.getSteamRating(),oldSteamRating,uri);
        }
        else if (mimeType.equals(HoursPlayedData.CONTENT_ITEM_TYPE)) {
          oldHoursPlayed = c.getString(2);
          contactOperation.updateHoursPlayed("" + friend.getHoursPlayed(),oldHoursPlayed,uri);
        }
        else if (mimeType.equals(HeadlineData.CONTENT_ITEM_TYPE)) {
          oldHeadline = c.getString(2);
          contactOperation.updateHeadline(friend.getHeadline(),oldHeadline,uri);
        }
        else if (mimeType.equals(StructuredPostal.CONTENT_ITEM_TYPE)) {
          int type = c.getInt(3);
          String label = c.getString(4);
          oldLocation = c.getString(2);
          if (type == StructuredPostal.TYPE_CUSTOM && label.equals(context.getString(R.string.label_location))) {
            oldLocation = c.getString(2);
            contactOperation.updateLocation(friend.getLocation(),oldLocation,uri);
          }
        }
        else if (mimeType.equals(StructuredName.CONTENT_ITEM_TYPE)) {
          oldRealName = c.getString(2);
          contactOperation.updateRealName(friend.getRealName(),oldRealName,uri);
        }
        else if (mimeType.equals(SummaryData.CONTENT_ITEM_TYPE)) {
          oldSummary = c.getString(2);
          contactOperation.updateSummary(friend.getSummary(),oldSummary,uri);
        }
        else if (mimeType.equals(MostPlayedGameData.CONTENT_ITEM_TYPE)) {
          oldMostPlayedGame = c.getString(2);
          contactOperation.updateMostPlayedGame(friend.getMostPlayedGame(),oldMostPlayedGame,uri);
        }
        else if (mimeType.equals(MostPlayedGameHoursData.CONTENT_ITEM_TYPE)) {
          oldMostPlayedGameHours = c.getString(2);
          contactOperation.updateMostPlayedGameHours("" + friend.getMostPlayedGameHours(),oldMostPlayedGameHours,uri);
        }
        else if (mimeType.equals(WebsiteData.CONTENT_ITEM_TYPE)) {
          oldWebsite = c.getString(2);
          contactOperation.updateWebsite(friend.getWebsite(),oldWebsite,uri);
        }
        else if (mimeType.equals(WebsiteTitleData.CONTENT_ITEM_TYPE)) {
          oldWebsiteTitle = c.getString(2);
          contactOperation.updateWebsiteTitle(friend.getWebsiteTitle(),oldWebsiteTitle,uri);
        }
        else if (mimeType.equals(PrimaryGroupData.CONTENT_ITEM_TYPE)) {
          oldPrimaryGroup = c.getString(2);
          contactOperation.updatePrimaryGroup(friend.getPrimaryGroup(),oldPrimaryGroup,uri);
        }
        else if (mimeType.equals(FavoriteGameData.CONTENT_ITEM_TYPE)) {
          oldFavoriteGame = c.getString(2);
          contactOperation.updateFavoriteGame(friend.getFavoriteGame(),oldFavoriteGame,uri);
        }
        else if (mimeType.equals(StatusData.CONTENT_ITEM_TYPE)) {
          oldStatus = c.getString(2);
          contactOperation.updateStatus(friend.getStateMessage(),oldStatus,uri);
          // Update the status icon too
          contactOperation.updateStatusIcon(friend.getStateMessage(),oldStatus,rawStatusId);
        }
        else if (mimeType.equals(GroupMembership.CONTENT_ITEM_TYPE)) {
          // The group never changes, so no update operation is required here. 
          // Just set the oldGroupMembership to a value so its not added to the group down below.
          oldGroupMembership = c.getLong(2);
        }
        else if (mimeType.equals(Photo.CONTENT_ITEM_TYPE)) {
          boolean timeExpired = false;
          try {
            long oldTimestamp = Long.parseLong(c.getString(12));
            long timestamp = Calendar.getInstance().getTimeInMillis();

            // Check that the photo is at least 1 day old

            long timeDifference = timestamp - oldTimestamp;
            if (timeDifference > 82800000) { // 23 hours in milliseconds
              timeExpired = true;
            }

          }
          catch (NumberFormatException e) {
            timeExpired = false;
          }

          if (timeExpired) {
            contactOperation.updatePhoto(friend.getPhoto(),uri);
          }

          // If it gets here, this contact already has a photo, so don't want to add a new photo!
          photoUpdated = true;
        }
      }
    }
    finally {
      c.close();
    }

    // For all unchanged values, add them from the <i>friend</i>
    if (oldUserID == -1) {
      contactOperation.addUserId(friend.getUserID());
    }
    if (oldUsername == null) {
      contactOperation.addUsername(friend.getUsername());
    }
    if (oldOnlineState == null) {
      contactOperation.addOnlineState(friend.getOnlineState());
    }
    if (oldCustomUrl == null) {
      contactOperation.addCustomUrl(friend.getCustomUrl());
    }
    if (oldMemberSince == null) {
      contactOperation.addMemberSince(friend.getMemberSince());
    }
    if (oldSteamRating == null) {
      contactOperation.addSteamRating("" + friend.getSteamRating());
    }
    if (oldHoursPlayed == null) {
      contactOperation.addHoursPlayed("" + friend.getHoursPlayed());
    }
    if (oldHeadline == null) {
      contactOperation.addHeadline(friend.getHeadline());
    }
    if (oldLocation == null) {
      contactOperation.addLocation(friend.getLocation());
    }
    if (oldRealName == null) {
      contactOperation.addRealName(friend.getRealName());
    }
    if (oldSummary == null) {
      contactOperation.addSummary(friend.getSummary());
    }
    if (oldMostPlayedGame == null) {
      contactOperation.addMostPlayedGame(friend.getMostPlayedGame());
    }
    if (oldMostPlayedGameHours == null) {
      contactOperation.addMostPlayedGameHours("" + friend.getMostPlayedGameHours());
    }
    if (oldWebsite == null) {
      contactOperation.addWebsite(friend.getWebsite());
    }
    if (oldWebsiteTitle == null) {
      contactOperation.addWebsiteTitle(friend.getWebsiteTitle());
    }
    if (oldPrimaryGroup == null) {
      contactOperation.addPrimaryGroup(friend.getPrimaryGroup());
    }
    if (oldFavoriteGame == null) {
      contactOperation.addFavoriteGame(friend.getFavoriteGame());
    }
    if (oldStatus == null) {
      contactOperation.addStatus(friend.getStateMessage());
      // Update the status icon too
      contactOperation.addStatusIcon(friend.getStateMessage(),rawStatusId);
    }

    if (oldGroupMembership == -1) {
      contactOperation.addGroupMembership(rawGroupId);
    }

    if (!photoUpdated) {
      contactOperation.addPhoto(friend.getPhoto());
    }

  }
}
